//+------------------------------------------------------------------+
//|                                                3LevelPattern.mq4 |
//+------------------------------------------------------------------+
#property copyright "Copyright  2011, TheXpert"
#property link      "theforexpert@gmail.com"

//   
extern int        Magic          = 23479269;
extern int        TimesToRepeat  = 3;
extern int        Slippage       = 3;

extern double     RiskPercentage = 0;
extern double     TradeLot       = 0;

extern int        SL             = 1000;
extern int        TP             = 0;

//  
// ,       .
extern string     ResultsFileName= "3LevelPatternResults.csv";
// ,     
extern string     NetFileName    = "3LevelPattern.ennf";
//  ,    . ..          LearnMonths 
extern int        LearnMonths    = 30;

//   
extern int        Sensors1       = 15;
extern int        Sensors2       = 10;
extern int        Sensors3       = 5;

extern int        Multiplier2    = 4;
extern int        Multiplier3    = 16;
extern double     HPFilterKoef   = 2.0;

//  
extern bool       TestRun        = false;
extern bool       AutoLearn      = false;

#include <EchoNet.mqh>
#include <DragTrade_v_1.0/DragTrade_Trading.mqh>

datetime LastTime;
datetime LastMonth;
bool FirstRun;
string _Symbol;
datetime StartTime;
int init()
{
   LastTime = 0;
   LastMonth = -1;
   FirstRun = true;
   _Symbol = Symbol();
   
   StartTime = TimeCurrent();

   return(0);
}

int start()
{
   if (Bars < 2*Sensors1) return;
   if (FirstRun)
   {
      OnFirstRun();
   }
   
   if (LastTime != Time[0])
   {
      LastTime = Time[0];
      OnNewBar();
   }
   else return (0);
   
   datetime startMonth = TimeMonth(TimeCurrent());
   if (LastMonth != startMonth)
   {
      //new month
      OnNewMonth();
      LastMonth = startMonth;
   }
   
   FirstRun = false;

   return(0);
}

int deinit()
{
   if (TestRun) LearnPatterns();
}

void OnFirstRun()
{
   Randomize(GetTickCount());
   
   SetLogName(ResultsFileName);
   ClearLog();
   
   if (AutoLearn) 
   {
      for (int i = 0; i < 24; i++)
      {
         FileDelete(NetFileName);
         FileDelete(ResultsFileName);
      }
   }
}

void OnNewBar()
{
   if (AutoLearn || TestRun) PreparePattern(); //     
   
   if (!TestRun)
   {
      if (AutoLearn && StartTime > TimeCurrent() - LearnMonths*30*24*60*60) return; //        

      double input[], output[];
      double out[1];
      double min, max;
      
      //  
      if (!PrepareInput(1, input, min, max)) return;
      
      // only to know output size.    
      if (!PrepareOutput(1 + Sensors1, output, min, max)) return;
      
      int inSize = ArraySize(input);
      int outSize = ArraySize(output);

      //    
      double DynamicID = GetTickCount();
      while(!RegisterUsage(DynamicID, inSize, outSize, true))
      {
         DynamicID = GetTickCount();
      }
   
      bool result = true;

      //   
      if (
            LoadNetFromFile(DynamicID, GetFullPath(NetFileName))
         )
      {
         //        output
         result = result && PropagatePattern(DynamicID, inSize, input, outSize, output);
      }
      else
      {
         Print("Load failed from ", GetFullPath(NetFileName));
         result = false;
      }
   
      //    
      UnregisterUsage(DynamicID);
      
      //   (    )
      double hi[], lo[];
      int size = (ArraySize(output))/2;
      ArrayResize(hi, size);
      ArrayResize(lo, size);
      
      for (int i = 0; i < size; i++)
      {
         lo[i] = output[i];
         hi[i] = output[i + size];
      }
      
      //     
      if (result) EnterMarket(hi, lo, min, max);
      
      DrawPatternLevel("Prediction", 1 - Sensors1, hi, lo, min, max, White, 5);
   }
}

//  ()   .      
void EnterMarket(double highs[], double lows[], double min, double max)
{
   double hi[], lo[];
   ArrayCopy(hi, highs);
   ArrayCopy(lo, lows);
   
   int size = ArraySize(hi);
   for (int i = 0; i < size; i++)
   {
      hi[i] = min + (max - min)*hi[i];
      lo[i] = min + (max - min)*lo[i];
   }
   
   int minPos = ArrayMinimum(lo);
   int maxPos = ArrayMaximum(hi);
   
   bool sellSignal = false, buySignal = false;
   bool closeSells = false, closeBuys = false;
   
   double ratio = (Bid - lo[minPos]);
   
   if (ratio > 0) ratio = (hi[maxPos] - Bid)/ratio;
   else 
   {
      ratio = 6;
   }
   
   if (ratio > 5) buySignal = true;
   if (ratio < 1/5) sellSignal = true;
   
   if (ratio < 1/2 && minPos < maxPos) closeBuys = true;
   if (ratio > 2 && minPos > maxPos) closeSells = true;
   
   if (closeBuys) CloseBuys(Magic, Slippage);
   if (closeSells) CloseSells(Magic, Slippage);
   
   
   if (buySignal)
   {
      if (GetOrdersCount(Magic, OP_BUY) == 0)
      {
         OpenBuy(Magic, TP, SL, TradeLot);
      }
   }
   
   if (sellSignal)
   {
      if (GetOrdersCount(Magic, OP_SELL) == 0)
      {
         OpenSell(Magic, TP, SL, TradeLot);
      }
   }
}

void OnNewMonth()
{
   //   
   if (!TestRun && AutoLearn) LearnPatterns();
}

//   
void LearnPatterns()
{
   //        
   if (AutoLearn && StartTime > TimeCurrent() - LearnMonths*30*24*60*60) return;

   int hFile = FileOpen(ResultsFileName, FILE_CSV | FILE_READ, " ");
   if (hFile == -1) return;
   
   Print("File opened");

   double input[], output[];
   double min, max;
   
   datetime date;
   
   //   ,    \
   if (!ReadPattern(hFile, date, input, output))
   {
      FileClose(hFile);
      Print("Pattern not read");
      return;
   }
   
   int inSize = ArraySize(input);
   int outSize = ArraySize(output);
   
   //  
   double DynamicID = GetTickCount();
   while(!RegisterUsage(DynamicID, inSize, outSize, true))
   {
      DynamicID = GetTickCount();
   }
   
   //    ,    
   //   .     
   if (!AddNeuronsToNet(DynamicID, 50, ETYPE_RSDN))
   {
      Print("AddNeuronsToNet failed");
   }

   if (!AddNeuronsToNet(DynamicID, 5, ETYPE_LINEAR))
   {
      Print("AddNeuronsToNet failed");
   }

   //   
   if (!AddRandomConnectionsToNet(DynamicID, 120))
   {
      Print("AddRandomConnectionsToNet failed");
   }

   //  
   if (!InitNetInputMatrix(DynamicID, inSize))
   {
      Print("InitNetInputMatrix failed");
   }   
   
   //    
   int count = 0;
   if (Time[0] < date + LearnMonths*60*60*24*30)
   {
      AddPatternsToNet(DynamicID, 1, inSize, input, outSize, output);
      count++;
   }

   while (ReadPattern(hFile, date, input, output))
   {
      if (Time[0] < date + LearnMonths*60*60*24*30)
      {
         AddPatternsToNet(DynamicID, 1, inSize, input, outSize, output);
         count++;
      }
   }

   //   
   if (count > 2)
   {
      int start = GetTickCount();
      Print("Starting Learn, count = ", count);
      if (LearnNet(DynamicID))
      {
         SaveNetToFile(DynamicID, GetFullPath(NetFileName));
      }
      else
      {
         FileDelete(NetFileName);
      }
      
      Print("Learn lasted ", (GetTickCount() - start)/1000, " seconds");
      Print("Should have been saved to -- ", GetFullPath(NetFileName));
   }
   else
   {
      Print("Not enough data, count = ", count);
   }
   
   // 
   UnregisterUsage(DynamicID);   

   FileClose(hFile);
}

//   .     -- .     
bool PrepareInput(int index, double& input[], double& min, double& max)
{
   double hi1[], lo1[];
   
   if (!FormPatternLevel(index, 1, Sensors1, hi1, lo1)) return (false);
   min = lo1[ArrayMinimum(lo1)];
   max = hi1[ArrayMaximum(hi1)];
   if (!NormalizeData(hi1, lo1, min, max)) return (false);
   
   
   double hi2[], lo2[];
   
   if (!FormPatternLevel(index, Multiplier2, Sensors2, hi2, lo2)) return (false);
   if (!NormalizeData(hi2, lo2, min, max)) return (false);


   double hi3[], lo3[];
   
   if (!FormPatternLevel(index, Multiplier3, Sensors3, hi3, lo3)) return (false);
   if (!NormalizeData(hi3, lo3, min, max)) return (false);
   
   DrawPatternLevel("0Level3", index, hi3, lo3, min, max, Yellow, 5);
   DrawPatternLevel("1Level2", index, hi2, lo2, min, max, Blue, 3);
   DrawPatternLevel("2Level1", index, hi1, lo1, min, max, Red, 1);

   int commonSize = 2 + 
         2*ArraySize(lo1) +
         2*ArraySize(lo2) +
         2*ArraySize(lo3);
         
   ArrayResize(input, commonSize);
   
   int pos = 0;

   input[pos] = min;                pos++;
   input[pos] = max;                pos++;
   ArrayCopy(input, lo1, pos, 0);   pos += ArraySize(lo1);
   ArrayCopy(input, hi1, pos, 0);   pos += ArraySize(hi1);
   ArrayCopy(input, lo2, pos, 0);   pos += ArraySize(lo2);
   ArrayCopy(input, hi2, pos, 0);   pos += ArraySize(hi2);
   ArrayCopy(input, lo3, pos, 0);   pos += ArraySize(lo3);
   ArrayCopy(input, hi3, pos, 0);
   
   return (true);
}

//     
bool PrepareOutput(int index, double& output[], double min, double max)
{
   double hi[], lo[];
   
   if (!FormPatternLevel(index - Sensors1, 1, Sensors1, hi, lo)) return (false);
   if (!NormalizeData(hi, lo, min, max)) return (false);
   
   if (TestRun) DrawPatternLevel("Output", index - Sensors1, hi, lo, min, max, White, 3);

   ArrayResize(output, 2*ArraySize(hi));

   int pos = 0;

   ArrayCopy(output, lo, pos, 0);   pos += ArraySize(lo);
   ArrayCopy(output, hi, pos, 0);

   return (true);
}

//    
void PreparePattern()
{
   double input[], output[];
   double min, max;
   
   if (!PrepareInput(Sensors1 + 1, input, min, max)) return;
   if (!PrepareOutput(Sensors1 + 1, output, min, max)) return;
   
   Log(Time[Sensors1 + 1]);
   LogArray(input);
   LogArray(output);
}

//   (  -  -    .  )        .
bool ReadPattern(int hFile, datetime& date, double& input[], double& output[])
{
   if (FileIsEnding(hFile)) return (false);
   
   date = FileReadNumber(hFile);
   if (date == 0) date = FileReadNumber(hFile);
   
   int inSize = FileReadNumber(hFile);
   if (inSize == 0) inSize = FileReadNumber(hFile);
   ArrayResize(input, inSize);
   
   string s;
   
   for (int i = 0; i < inSize; i++)
   {
      input[i] = FileReadNumber(hFile);
   }
   
   int outSize = FileReadNumber(hFile);
   if (outSize == 0) outSize = FileReadNumber(hFile);
   ArrayResize(output, outSize);
   
   for (i = 0; i < outSize; i++)
   {
      output[i] = FileReadNumber(hFile);
   }
   
   if (inSize <= 0 || outSize <= 0) return (false);
   
   return (true);
}

void DrawPatternLevel(string name, int index, double hi[], double lo[], double min, double max, color clr, int width)
{
   if (IsTesting() && !IsVisualMode()) return;
   
   int size = ArraySize(hi);
   if (size < 1) return;
   if (ArraySize(lo) != size) return;
   
   for (int i = 0; i < size; i++)
   {
      CreateHBar(name + i, min + lo[i]*(max - min), min + hi[i]*(max - min), GetTime(index + i), clr, width);
   }
}

void CreateHBar(string name, double lo, double hi, datetime time, color clr, int width)
{
   ObjectCreate(name, OBJ_TREND, 0, time, lo, time, hi);
   ObjectSet(name, OBJPROP_COLOR, clr);
   ObjectSet(name, OBJPROP_TIME1, time);
   ObjectSet(name, OBJPROP_TIME2, time);
   ObjectSet(name, OBJPROP_PRICE1, lo);
   ObjectSet(name, OBJPROP_PRICE2, hi);
   ObjectSet(name, OBJPROP_RAY, false);
   ObjectSet(name, OBJPROP_WIDTH, width);
}

datetime GetTime(int index)
{
   if (index >= 0) return (Time[index]);
   return (Time[0] - index*60*Period());
}


//********************************************************

string LogFileName;

void SetLogName(string fileName)
{
   LogFileName = fileName;
}

void LogArray(double array[])
{
   int size = ArraySize(array);
   string s = size + " ";
   for (int i = 0; i < size; i++)
   {
      s = s + array[i] + " ";
   }
   Log(s);
}

void Log(string s)
{
   int hFile = FileOpen(LogFileName, FILE_READ | FILE_WRITE | FILE_CSV, " ");
   if (hFile == -1) return;
   
   if (!FileSeek(hFile, 0, SEEK_END)) 
   {
      FileClose(hFile); 
      return;
   }
   
   FileWrite(hFile, s);
   FileClose(hFile);
}

void ClearLog()
{
   int hFile = FileOpen(ResultsFileName, FILE_WRITE | FILE_CSV, " ");
   if (hFile == -1) return;
   
   FileClose(hFile);
}

//********************************************************

void HPFilter(int nobs, int lambda, double& input[], double& output[])
{
   double a[], b[], c[];
   ArrayResize(a,nobs);
   ArrayResize(b,nobs);
   ArrayResize(c,nobs);

   double H1 = 0.0, H2 = 0.0, H3 = 0.0, H4 = 0.0, H5 = 0.0, HH1 = 0.0, HH2 = 0.0, HH3 = 0.0, HH5 = 0.0, HB = 0.0, HC = 0.0, Z = 0.0;
   ArrayResize(output, ArraySize(input));
   ArrayCopy(output, input);

   a[0] = 1.0 + lambda;
   b[0] = -2.0*lambda;
   c[0] = lambda;
   
   double _a = 6.0*lambda + 1.0;
   double _b = -4.0*lambda;
   
   for (int i = 1; i < nobs - 2; i++)
   {
      a[i] = _a;
      b[i]= _b;
      c[i]=lambda;
   }
   
   a[1]= 5.0*lambda + 1;
   a[nobs - 1] = 1.0 + lambda;
   a[nobs - 2] = 5.0*lambda + 1.0;
   b[nobs - 2] = -2.0*lambda;
   b[nobs - 1] = 0.0;
   c[nobs - 2] = 0.0;
   c[nobs - 1] = 0.0;

   //Forward
   for (i = 0; i < nobs; i++)
   {
      Z=a[i] - H4*H1 - HH5*HH2;
      HB = b[i];
      HH1 = H1;
      H1 = (HB - H4*H2)/Z;
      b[i] = H1;
      HC = c[i];
      HH2 = H2;
      H2 = HC/Z;
      c[i] = H2;
      a[i] = (output[i] - HH3*HH5 - H3*H4)/Z;
      HH3 = H3;
      H3 = a[i];
      H4 = HB - H5*HH1;
      HH5 = H5;
      H5 = HC;
   }

   //Backward 
   H2 = 0;
   H1 = a[nobs - 1];
   output[nobs - 1]=H1;
   for (i = nobs - 2; i >= 0; i--)
   {
      output[i] = a[i] - b[i]*H1 - c[i]*H2;
      H2 = H1;
      H1 = output[i];
   }
}

bool FormPatternLevel(int index, int multiplier, int size, double& hiValues[], double& loValues[])
{
   if (multiplier < 1) return (false);
   if (size < 1) return (false);
   if (index < 0 || index + multiplier*size >= Bars) return (false);

   double hi[], lo[];
   ArrayResize(hi, size);
   ArrayResize(lo, size);
   
   for (int i = 0; i < size; i++)
   {
      hi[i] = 0;
      lo[i] = 1000000;
      
      for (int j = 0; j < multiplier; j++)
      {
         double value = High[index + multiplier*i + j];
         if (value > hi[i]) hi[i] = value;

         value = Low[index + multiplier*i + j];
         if (value < lo[i]) lo[i] = value;
      }
   }

   HPFilter(size, HPFilterKoef, hi, hiValues);
   HPFilter(size, HPFilterKoef, lo, loValues);
   
   return (true);
}

bool NormalizeData(double& hi[], double& lo[], double min, double max)
{
   int size = ArraySize(hi);
   if (ArraySize(lo) != size) return (false);

   double scale = max - min;
   if (scale <= 0) return (false);
   
   for (int i = 0; i < size; i++)
   {
      hi[i] = (hi[i] - min)/scale;
      lo[i] = (lo[i] - min)/scale;
   }
   
   return (true);
}

string GetFullPath(string name)
{
   static string result;
   result = TerminalPath() + "/Experts/Files/" + name;
   if (IsTesting()) result = TerminalPath() + "/Tester/Files/" + name;
   
   return (result);
}